
'''
This file is automatically generated; Do not edit it. :)
'''
IS_FINAL = True
VERSION = '39.1.0'
BRANCH_NICK = 'release-39'
REVISION_ID = '81173dd2f0100a1ea7314ab240c9799aca2b0a35'
REVNO = '14273'
